/*****************************************************************************/
/*
                               CgiVar.c

These functions provides a simple, uniform mechanism for a C Language script to
access CGI variable values transparently in the WASD CGI and CGIplus
environments, as well as OSU (once DCL symbols have been created for the CGI
variables) and "standard" CGI (e.g. Netscape FastTrack) environments.  CGI
variables names are assumed to be in upper-case.

It is completely self-contained. Simply call the CgiVar() function with the CGI
variable name value to be accessed and if that CGI variable exists a pointer to
the value will be returned.  Non-existant variables return CGIPLUS_NONE (see
below).  The values pointed to should NOT be modified in any way (copy the
value if this is required).  There is no need to call the other functions in
this module. 

In the CGIplus environment supplying an empty variable name (i.e. "") will
cause the function to block  until the arrival of a request, upon which it just
returns a CGIPLUS_NONE pointer, allowing start-of-request synchronization. 
This will also release allocated memory and reset in preparation for reading
the next request, an alternative to calling with a NULL parameter at the end of
request processing (see below).

For the CGIplus environment supplying a wildcard variable name (i.e. "*")
returns each CGIplus variable string (line from CGIPLUSIN) in turn, with no
more being indicated by a CGIPLUS_NONE pointer.

At the end of a CGIplus request call the function with a NULL parameter to
release allocated memory and reset ready for processing the next request's
variables.  It is not necessary to do this if calling with an empty parameter
to synchronize on the start of a request (see above).

The CGIplus functionality works by initially reading all lines from
CGIPLUSIN until the terminating blank line is encountered. These lines are
stored as a series of contiguous null-terminated strings, terminated by an
empty string, in allocated memory. These strings are searched for the required
variable name and a pointer to its value returned if found.

The value CGIPLUS_NONE shold be set according to the requirements of the
application the function is included within.  If calling functions can handle
NULL pointers being returned then set the value to NULL.  If non-NULL pointers
are prefered then set the value to an empty string.  The first example would
return an empty string, the second a NULL pointer.

  #define CGIPLUS_NONE ""
  #define CGIPLUS_NONE NULL


INCLUDING IN YOUR CODE
----------------------
Include the CgiVar() function code from the standard WASD location
(HT_ROOT:[SRC.CGIPLUS]) if CGIVAR_INCLUDE has not been defined.
If defined then it must be the #include location.
The following is a suggested method for doing this.

  #ifndef CGIVAR_INCLUDE
  #   include "../cgiplus/cgivar.c"
  #else
  #   include CGIVAR_INCLUDE
  #endif

Debug statements could be tuned on or off by include the 'Debug' boolean in a
macro such as the following (see WASD scripting example for further example).

  #define CGIPLUS_DEBUG (Debug)


VERSION HISTORY (update SoftwareID as well!)
---------------
24-APR-1999  MGD  v1.3.0, standard CGI environment processing (e.g. FastTrack)
14-MAR-1999  MGD  v1.2.0, refine reading CGIPLUSIN stream reducing overhead
21-JUN-1998  MGD  v1.1.0, increase processing efficiency by reducing the
                          number of realloc()s required and introducing the
                          length of each variable improving search behaviour
01-JUL-1997  MGD  v1.0.0 initial
*/
/*****************************************************************************/

#ifdef __ALPHA
   char CgiVarSoftwareID [] = "CGIVAR AXP-1.3.0";
#else
   char CgiVarSoftwareID [] = "CGIVAR VAX-1.3.0";
#endif

#ifndef CGIPLUS_DEBUG
#  define CGIPLUS_DEBUG 0
#endif
#ifndef CGIPLUS_NONE
#   define CGIPLUS_NONE ""
/** #define CGIPLUS_NONE NULL **/
#endif

#ifndef OSU_ENVIRONMENT
#  define OSU_ENVIRONMENT 1
#endif

#ifndef STD_CGI_ENVIRONMENT
#  define STD_CGI_ENVIRONMENT 1
#endif

#define CGIPLUS_LINE_SIZE 1024
#define CGIPLUS_CHUNK 2048

/* prototypes */
char* CgiVarGetKey (char*);
char* CgiVarGetForm (char*);
char* CgiVarGetPathTranslated ();
int CgiVarUrlDecode (char*);

/*****************************************************************************/
/*
*/

char* CgiVar (char *VarName)

{
   static int  CGIplusEnvironment = 0,
               InitializeEnvironment = 1,
               LoadVariables = 1,
               VarPtrSize = 0,
               WWWprefix = 0;

   static char  *NextPtr = NULL,
                *VarPtr = NULL;

   register char  *cptr, *sptr, *zptr;

   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (CGIPLUS_DEBUG) fprintf (stdout, "CgiVar() |%s|\n", VarName);

   if (InitializeEnvironment)
   {
      WWWprefix = (getenv ("WWW_GATEWAY_INTERFACE") != NULL);
      CGIplusEnvironment = (getenv ("CGIPLUSEOF") != NULL);
      InitializeEnvironment = 0;
   }

   if (!CGIplusEnvironment)
   {
      /***************************/
      /* non-CGIplus environment */
      /***************************/

      if (WWWprefix)
      {
         /* WASD, OSU (with symbols generated), probably Purveyor */
         if ((cptr = getenv(VarName)) == NULL)
         {
            if (CGIPLUS_DEBUG) fprintf (stdout, "|%s|=CGIPLUS_NONE\n", VarName);
            return (CGIPLUS_NONE);
         }
         else
         {
            if (CGIPLUS_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, cptr);
            return (cptr);
         }
      }

      /* at least Netscape FastTrack */
      if (!memcmp (VarName+4, "KEY_", 4))
         return (CgiVarGetKey (VarName+4));

      if (!memcmp (VarName+4, "FORM_", 5))
         return (CgiVarGetForm (VarName+9));

      if (!strcmp (VarName+4, "PATH_TRANSLATED"))
         return (CgiVarGetPathTranslated ());

      if ((cptr = getenv(VarName+4)) == NULL)
      {
         if (CGIPLUS_DEBUG) fprintf (stdout, "|%s|=CGIPLUS_NONE\n", VarName);
         return (CGIPLUS_NONE);
      }
      else
      {
         if (CGIPLUS_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, cptr);
         return (cptr);
      }
   }

   /***********************/
   /* CGIplus environment */
   /***********************/

   if (!VarName[0])
   {
      /* end of request, reinitialize */
      LoadVariables = 1;
      NextPtr = NULL;
   }

   if (LoadVariables)
   {
      /* read lines containing CGIplus variables from <CGIPLUSIN> */
      static FILE  *CgiPlusIn = NULL;

      int  CurrentVarLength;
      char  *LengthPtr;
      char  Line [CGIPLUS_LINE_SIZE];
                                                 
      LoadVariables = 0;

      if (CgiPlusIn == NULL)
         if ((CgiPlusIn = fopen (getenv("CGIPLUSIN"), "r")) == NULL)
            exit (vaxc$errno);

      CurrentVarLength = 0;
      zptr = VarPtr + VarPtrSize;
      while (fgets (Line, sizeof(Line), CgiPlusIn) != NULL)
      {
         /* ignore any non-alphabetic line */
         if (Line[0] != '\n' && !isalpha(Line[0])) continue;
         /* point to the next name=value storage in the variable buffer */
         LengthPtr = VarPtr + CurrentVarLength;
         sptr = LengthPtr + sizeof(unsigned short);
         for (cptr = Line; *cptr; *sptr++ = *cptr++)
         {
            if (sptr < zptr) continue;
            /* out of storage (or first use) get more (or some) memory */
            if ((VarPtr = realloc (VarPtr, VarPtrSize+CGIPLUS_CHUNK)) == NULL)
               exit (vaxc$errno);
            VarPtrSize += CGIPLUS_CHUNK;
            /* recalculate pointers */
            LengthPtr = VarPtr + CurrentVarLength;
            sptr = LengthPtr + sizeof(unsigned short) + (cptr - Line);
            zptr = VarPtr + VarPtrSize;
         }
         /* remove the trailing newline */
         *--sptr = '\0';
         /* insert the length of this name=value pair immediately before it */
         *((unsigned short*)LengthPtr) = Length = cptr - Line;
         /* adjust the value of the current variable storage space used */
         CurrentVarLength += Length + sizeof(unsigned short);
         /* first empty line signals the end of CGIplus variables */
         if (Line[0] == '\n') break;
      }

      if (VarPtr == NULL) return (CGIPLUS_NONE);

      if (CGIPLUS_DEBUG)
      {
         fprintf (stdout, "VarPtr: %d VarPtrSize: %d CurrentVarLength: %d\n",
                  VarPtr, VarPtrSize, CurrentVarLength);
         sptr = VarPtr;
         for (sptr = VarPtr;
              Length = *((unsigned short*)sptr);
              sptr += sizeof(unsigned short) + Length)
            fprintf (stdout, "%3d |%s|\n", Length, sptr+sizeof(unsigned short));
      }
   }

   /* if just waiting for next request return now it's arrived */
   if (!VarName[0]) return (CGIPLUS_NONE);

   if (VarName[0] == '*')
   {
      /* return each stored line one at a time */
      if (NextPtr == NULL) NextPtr = VarPtr;
      if ((Length = *((unsigned short*)(sptr = NextPtr))) == 1)
      {
         NextPtr = NULL;
         if (CGIPLUS_DEBUG) fprintf (stdout, "|%s|=CGIPLUS_NONE\n", VarName);
         return (CGIPLUS_NONE);
      }
      sptr += sizeof(unsigned short);
      NextPtr = sptr + Length;
      if (CGIPLUS_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, sptr);
      return (sptr);
   }

   /* search CGIplus variable strings for required variable */
   for (sptr = VarPtr;
        Length = *((unsigned short*)sptr);
        sptr = zptr + Length)
   {
      /* simple comparison between supplied and in-string variable names */
      zptr = sptr = sptr + sizeof(unsigned short);
      cptr = VarName;
      while (*cptr && *sptr && *sptr != '=')
      {
         if (toupper(*cptr) != toupper(*sptr)) break;
         cptr++;
         sptr++;
      }
      if (*cptr || *sptr != '=') continue;
      /* found, return a pointer to the value */
      if (CGIPLUS_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, sptr+1);
      return (sptr+1);
   }
   /* not found */
   if (CGIPLUS_DEBUG) fprintf (stdout, "|%s|=CGIPLUS_NONE\n", VarName);
   return (CGIPLUS_NONE);
}

#if STD_CGI_ENVIRONMENT

/*****************************************************************************/
/*
For "standard" CGI environment (e.g. Netscape FastTrack) derive the
"KEY_COUNT", "KEY_1" ... "KEY_n" variables from "QUERY_STRING".
*/

char* CgiVarGetKey (char *VarName)

{
   static char  *QueryStringPtr = NULL;

   register char  *aptr, *cptr, *sptr;

   int  KeyCount,
        VarKeyCount;

   /*********/
   /* begin */
   /*********/

   if (CGIPLUS_DEBUG) fprintf (stdout, "CgiVarGetKey() |%s|\n", VarName);

   if (!VarName[0]) return (CGIPLUS_NONE);

   if ((cptr = QueryStringPtr) == NULL)
      if ((QueryStringPtr = cptr = getenv ("QUERY_STRING")) == NULL)
         return (CGIPLUS_NONE);

   if (!strcmp (VarName, "KEY_COUNT"))
   {
      KeyCount = 0;
      for (sptr = cptr; *sptr; sptr++)
      {
         if (*sptr == '=') return (CGIPLUS_NONE);
         if (*sptr == '+') KeyCount++;
      }
      if (sptr == cptr) return ("0");
      KeyCount++;
      sprintf (aptr = calloc (1, 16), "%d", KeyCount);
      if (CGIPLUS_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, aptr);
      return (aptr);
   }
 
   VarKeyCount = atoi (VarName+4);
   KeyCount = 1;
   while (*cptr && VarKeyCount)
   {
      sptr = cptr;
      while (*cptr && *cptr != '+')
      {
         if (*cptr == '=') return (CGIPLUS_NONE);
         cptr++;
      }
      if (KeyCount == VarKeyCount)
      {
         aptr = calloc (1, cptr - sptr + 1);
         memcpy (aptr, sptr, cptr-sptr);
         aptr[cptr-sptr] = '\0';
         CgiVarUrlDecode (aptr);
         if (CGIPLUS_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, aptr);
         return (aptr);
      }
      KeyCount++;
   }

   if (CGIPLUS_DEBUG) fprintf (stdout, "|%s|=CGIPLUS_NONE\n", VarName);
   return (CGIPLUS_NONE);
}

/*****************************************************************************/
/*
For "standard" CGI environment (e.g. Netscape FastTrack) derive the
"FORM_xxxxx" variables from "QUERY_STRING".
*/

char* CgiVarGetForm (char *VarName)
   
{
   static char  *QueryStringPtr = NULL;

   register char  *aptr, *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (CGIPLUS_DEBUG) fprintf (stdout, "CgiVarGetForm() |%s|\n", VarName);

   if ((cptr = QueryStringPtr) == NULL)
      if ((QueryStringPtr = cptr = getenv ("QUERY_STRING")) == NULL)
         return (CGIPLUS_NONE);

   while (*cptr)
   {
      sptr = cptr;
      while (*cptr && *cptr != '=') cptr++;
      if (!*cptr) return (CGIPLUS_NONE);
      *cptr = '\0';
      aptr = VarName;
      while (*aptr && *sptr && toupper(*aptr) == toupper(*sptr))
      {
         aptr++;
         sptr++;
      }
      if (*aptr || *sptr)
      {
         *cptr++ = '=';
         while (*cptr && *cptr != '&') cptr++;
         if (*cptr) cptr++;
         continue;
      }
      *cptr++ = '=';

      sptr = cptr;
      while (*cptr && *cptr != '&') cptr++;
      aptr = calloc (1, cptr - sptr + 1);
      memcpy (aptr, sptr, cptr-sptr);
      aptr[cptr-sptr] = '\0';
      CgiVarUrlDecode (aptr);
      if (CGIPLUS_DEBUG) fprintf (stdout, "|%s=%s|\n", VarName, aptr);
      return (aptr);
   }

   if (CGIPLUS_DEBUG) fprintf (stdout, "|%s|=CGIPLUS_NONE\n", VarName);
   return (CGIPLUS_NONE);
}

/*****************************************************************************/
/*
For "standard" CGI environment (e.g. Netscape FastTrack) get the
"PATH_TRANSLATED" variable and check if it's in Unix-style or VMS-style format. 
Convert to a VMS specification as necessary.
*/

char* CgiVarGetPathTranslated ()
   
{
   register char  *aptr, *cptr, *sptr;

   int  Length;
   char  PathTranslated [256];

   /*********/
   /* begin */
   /*********/

   if (CGIPLUS_DEBUG) fprintf (stdout, "CgiVarGetPathTranslated()\n");

   if ((aptr = cptr = getenv ("PATH_TRANSLATED")) == NULL)
      return (CGIPLUS_NONE);

   if (*cptr == '/')
   {
      /* unix-style specification */
      sptr = PathTranslated;
      if (*cptr == '/') cptr++;
      while (*cptr && *cptr != '/') *sptr++ = toupper(*cptr++);
      if (*cptr)
      {
         cptr++;
         *sptr++ = ':';
         *sptr++ = '[';
      }
      while (*cptr) *sptr++ = toupper(*cptr++);
      *sptr-- = '\0';
      while (sptr > PathTranslated && *sptr != '/') sptr--;
      if (*sptr == '/')
      {
         *sptr-- = ']';
         while (sptr >= PathTranslated)
         {
            if (*sptr == '/') *sptr = '.';
            if (*(unsigned long*)sptr == '....')
               memmove (sptr, sptr+1, strlen(sptr)); 
            sptr--;
         }
      }
      aptr = calloc (1, Length = strlen(PathTranslated) + 1);
      memcpy (aptr, PathTranslated, Length);
   }

   if (CGIPLUS_DEBUG) fprintf (stdout, "|PATH_TRANSLATED=%s|\n", aptr);
   return (aptr);
}

/*****************************************************************************/
/*
Decode URL-encoded string.  Resultant string is always the same size or smaller
so it can be done in-situ!  When processing in the "standard" CGI environment.
*/

int CgiVarUrlDecode (char *String)

{
   register char  c;
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "UrlDecode() |%s|\n", String);

   cptr = sptr = String;
   while (*cptr)
   {
      switch (*cptr)
      {
         case '+' :

            *sptr++ = ' ';
            cptr++;

            continue;

         case '%' :

            cptr++;
            c = 0;
            if (*cptr >= '0' && *cptr <= '9')
               c = (*cptr - (int)'0') << 4;
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               c = (toupper(*cptr) - (int)'A' + 10) << 4;
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            if (*cptr >= '0' && *cptr <= '9')
               c += (*cptr - (int)'0');
            else
            if (toupper(*cptr) >= 'A' && toupper(*cptr) <= 'F')
               c += (toupper(*cptr) - (int)'A' + 10);
            else
            {
               *String = '\0';
               return (-1);
            }
            if (*cptr) cptr++;
            *sptr++ = c;

            continue;

         default :
            
            *sptr++ = *cptr++;

      }
   }
   *sptr = '\0';

   if (Debug) fprintf (stdout, "|%s|\n", String);
   return (sptr-String);
}

#endif /* STD_CGI_ENVIRONMENT */

/*****************************************************************************/

